unit ScrolCan;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls;

type
  TScrollCanvas = class(TCustomControl)
  private
    { Private declarations }
    FPen: HPen;
    FBrush: HBrush;
    FRedraw: Boolean;
    BackBit: TBitmap;
    BackBrush: TBrush;
    ClassName: String;
    FOldWidth: Word;
    FOldHeight: Word;
    FPaintOnDrags: Boolean;
    FOnPaint: TNotifyEvent;
    FOnResize: TNotifyEvent;
    FHorzPos: Integer;
    FAreaWidth: Integer;
    FAreaMinHorz: Integer;
    FVertPos: Integer;
    FAreaHeight: Integer;
    FAreaMinVert: Integer;
    procedure SetHorzPos(Value: Integer);
    procedure SetVertPos(Value: Integer);
    procedure SetAreaWidth(Value: Integer);
    procedure SetAreaHeight(Value: Integer);
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  protected
    { Protected declarations }
    procedure WMVScroll(var Message: TWMScroll); message WM_VSCROLL;
    procedure WMHScroll(var Message: TWMScroll); message WM_HSCROLL;
    procedure WMSetRedraw(var Message: TWMSetRedraw); message WM_SETREDRAW;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
  public
    { Public declarations }
    procedure RecalcBars;
    property Canvas;
    property Redraw: Boolean read FRedraw write FRedraw;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure Paint; override;
    procedure PaintWindow(DC: HDC); override;
    procedure Resize;
  published
    { Published declarations }
    property Align;
    property Ctl3D;
    property Font;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property ParentCtl3D;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyUp;
    property OnKeyPress;
    property Visible;
    property Enabled;
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
    property OnResize: TNotifyEvent read FOnResize write FOnResize;
    property PaintOnDrags: Boolean read FPaintOnDrags write FPaintOnDrags;
    property HorzPos: Integer read FHorzPos write SetHorzPos;
    property VertPos: Integer read FVertPos write SetVertPos;
    property AreaWidth: Integer read FAreaWidth write SetAreaWidth default 400;
    property AreaHeight: Integer read FAreaHeight write SetAreaHeight default 400;
  end;

procedure Register;

implementation

{$R SCROLCAN.RES}

constructor TScrollCanvas.Create(AOwner: TComponent);
var
  LogBrush: TLOGBRUSH;
begin
  inherited Create(AOwner);
  Width := 200;
  Height := 200;
  BackBit := TBitmap.Create;
  BackBit.Handle := LoadBitmap(HInstance, 'CHECKER');
  BackBrush := TBrush.Create;
  BackBrush.Bitmap := BackBit;
  Brush.Assign(BackBrush);
  FAreaWidth := 400;
  FAreaMinHorz := 0;
  FAreaHeight := 400;
  FAreaMinVert := 0;
  Redraw := True;
  LogBrush.lbStyle := BS_HOLLOW;
  FBrush := CreateBrushIndirect(LogBrush);
  FPen := CreatePen(PS_SOLID, 0, COLOR_WINDOWFRAME);
end;

destructor TScrollCanvas.Destroy;
begin
  inherited Destroy;
end;

procedure TScrollCanvas.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
{  Params.WindowClass.hbrBackground := BackBrush.Handle;}
  Params.Style := Params.Style or WS_HSCROLL or WS_VSCROLL;
end;

procedure TScrollCanvas.Paint;
begin
  inherited Paint;
  if Assigned (FOnPaint) and Redraw then
    FOnPaint(Self);
end;

procedure TScrollCanvas.PaintWindow(DC: HDC);
var
  OldPen: HPen;
  OldBrush: HBrush;
begin
  if Redraw then
  begin
    inherited PaintWindow(DC);
    OldPen := SelectObject(DC, FPen);
    OldBrush := SelectObject(DC, FBrush);
    Rectangle(DC, 0, 0, ClientWidth, ClientHeight);
    SelectObject(DC, OldPen);
    SelectObject(DC, OldBrush);
  end;
end;

procedure TScrollCanvas.Resize;
begin
  if Assigned(FOnResize) then
    FOnResize(Self);
  RecalcBars;
end;

procedure TScrollCanvas.RecalcBars;
var
  ShowHorz, ShowVert: Boolean;
  HSHeight, VSWidth: Integer;
begin
  ShowHorz := False;
  ShowVert := False;
  HSHeight := GetSystemMetrics(SM_CYHSCROLL);
  VSWidth := GetSystemMetrics(SM_CXVSCROLL);
  if FAreaHeight > Height then
    ShowVert := True;
  if FAreaWidth > Width then
    ShowHorz := True;
  if ShowHorz and (FAreaHeight > Height - HSHeight) then
    ShowVert := True;
  if ShowVert and (FAreaWidth > Width - VSWidth) then
    ShowHorz := True;

  if FVertPos > AreaHeight - ClientHeight then FVertPos := AreaHeight - ClientHeight;
  if FVertPos < 0 then FVertPos := 0;
  if FHorzPos > AreaWidth - ClientWidth then FHorzPos := AreaWidth - ClientWidth;
  if FHorzPos < 0 then FHorzPos := 0;
  if ShowHorz then
  begin
    SetScrollRange(Handle, SB_HORZ, FAreaMinHorz, FAreaWidth - ClientWidth, False);
    SetScrollPos(Handle, SB_HORZ, FHorzPos, False);
  end;
  ShowScrollBar(Handle, SB_HORZ, ShowHorz);
  if ShowVert then
  begin
    SetScrollRange(Handle, SB_VERT, FAreaMinVert, FAreaHeight - ClientHeight, False);
    SetScrollPos(Handle, SB_VERT, FVertPos, False);
  end;
  ShowScrollBar(Handle, SB_VERT, ShowVert);
end;

procedure TScrollCanvas.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
end;

procedure TScrollCanvas.WMHScroll(var Message: TWMScroll);
var
  OldPos: Integer;
  ScrollRect: TRect;
begin
  OldPos := FHorzPos;
  case Message.ScrollCode of
   SB_BOTTOM:        FHorzPos := AreaWidth;
   SB_LINEDOWN:      FHorzPos := FHorzPos + 3;
   SB_LINEUP:	     FHorzPos := FHorzPos - 3;
   SB_PAGEDOWN:	     FHorzPos := FHorzPos + ClientWidth div 2;
   SB_PAGEUP:        FHorzPos := FHorzPos - ClientWidth div 2;
   SB_THUMBPOSITION: FHorzPos := Message.Pos;
   SB_THUMBTRACK:    FHorzPos := Message.Pos;
   SB_TOP:           FHorzPos := 0;
  end;
  if FHorzPos > AreaWidth - ClientWidth then FHorzPos := AreaWidth - ClientWidth;
  if FHorzPos < 0 then FHorzPos := 0;
  if OldPos <> FHorzPos then
  begin
    SetScrollPos(Handle, SB_HORZ, FHorzPos, True);
    if Redraw then
    begin
      ScrollRect := Rect(1, 1, ClientWidth - 1, ClientHeight - 1);
      ScrollWindow(Handle, OldPos - FHorzPos, 0, nil, @ScrollRect);
      UpdateWindow(Handle);
    end;
  end;
  inherited;
end;

procedure TScrollCanvas.WMVScroll(var Message: TWMScroll);
var
  OldPos: Integer;
  ScrollRect: TRect;
begin
  OldPos := FVertPos;
  case Message.ScrollCode of
   SB_BOTTOM:        FVertPos := AreaHeight;
   SB_LINEDOWN:      FVertPos := FVertPos + 3;
   SB_LINEUP:	     FVertPos := FVertPos - 3;
   SB_PAGEDOWN:	     FVertPos := FVertPos + ClientHeight div 2;
   SB_PAGEUP:        FVertPos := FVertPos - ClientHeight div 2;
   SB_THUMBPOSITION: FVertPos := Message.Pos;
   SB_THUMBTRACK:    FVertPos := Message.Pos;
   SB_TOP:           FVertPos := 0;
  end;
  if FVertPos > AreaHeight - ClientHeight then FVertPos := AreaHeight - ClientHeight;
  if FVertPos < 0 then FVertPos := 0;
  if OldPos <> FVertPos then
  begin
    SetScrollPos(Handle, SB_VERT, FVertPos, True);
    if Redraw then
    begin
      ScrollRect := Rect(1, 1, ClientWidth - 1, ClientHeight - 1);
      ScrollWindow(Handle, 0, OldPos - FVertPos, nil, @ScrollRect);
      UpdateWindow(Handle);
    end;
  end;
  inherited;
end;

procedure TScrollCanvas.WMSetRedraw(var Message: TWMSetRedraw);
begin
  Redraw := Boolean(Message.Redraw);
  inherited;
end;

procedure TScrollCanvas.WMSize(var Message: TWMSize);
begin
  inherited;
  if (FOldHeight <> Message.Height) or (FOldWidth <> Message.Width) then
    Resize;
  FOldHeight := Message.Height;
  FOldWidth := Message.Width;
end;

procedure TScrollCanvas.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  UnrealizeObject(BackBrush.Handle);
  SetBrushOrg(Message.DC, FVertPos mod 2, FHorzPos mod 2);
  SelectObject(Message.DC, BackBrush.Handle);
  Rectangle(Message.DC, 0, 0, ClientWidth, ClientHeight);
end;

procedure TScrollCanvas.SetHorzPos(Value: Integer);
var
  Msg: TWMScroll;
begin
  if Value <> FHorzPos then
  begin
    Msg.Msg := WM_HSCROLL;
    Msg.ScrollCode := SB_THUMBPOSITION;
    Msg.Pos := Value;
    Msg.ScrollBar := 0;
    WMHScroll(Msg);
  end;
end;

procedure TScrollCanvas.SetVertPos(Value: Integer);
var
  Msg: TWMScroll;
begin
  if Value <> FVertPos then
  begin
    Msg.Msg := WM_VSCROLL;
    Msg.ScrollCode := SB_THUMBPOSITION;
    Msg.Pos := Value;
    Msg.ScrollBar := 0;
    WMVScroll(Msg);
  end;
end;

procedure TScrollCanvas.SetAreaWidth(Value: Integer);
begin
  if Value <> FAreaWidth then
  begin
    FAreaMinHorz := 0;
    FAreaWidth := Value;
  end;
end;

procedure TScrollCanvas.SetAreaHeight(Value: Integer);
begin
  if Value <> FAreaHeight then
  begin
    FAreaMinVert := 0;
    FAreaHeight := Value;
  end;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TScrollCanvas]);
end;

end.
